//
//  EVCK.java
//  EVCK
//
//  Created by dario on 13/04/05.
//  Copyright 2005 __MyCompanyName__. All rights reserved.
//

import javax.imageio.ImageIO;
import java.awt.image.*;
import java.awt.*;
import java.io.*;

/**
 * Implements a 2 out of 2 threshold Extended Visual Cryptography Scheme 
*/
public class EVCS {
	//basis matrix to generate matrix collections for encoding
	private BasisMatrix[] S;
	//images
	private BufferedImage img1, img2, simg; //original images
	private BufferedImage share_img1, share_img2; // share images
	private int m=4; //expansion factor

	/**
	 * Constructor method throws a NotComplianSizeException if the input 
	 * images don't have the same size.
	 * @param si the secret image file name
	 * @param im1 user 1's innocent image
	 * @param im2 user 2's innocent image
	 */
	public EVCS(String si, String im1, String im2) throws NotCompliantSizeException
	{	//allocate the 2^n basis matrices braces
		S = new BasisMatrix[8];
		//create every basis matrix
		populateBasisMatrices();
		//load image files
		try
		{	img1 = ImageIO.read(new File(im1));
			img2 = ImageIO.read(new File(im2));
			simg = ImageIO.read(new File(si));
		}
		catch(IOException e)
		{	System.out.println("Errore nella lettura dei file: verificarne l'esistenza.");
			System.exit(1);
		}
		
		//CONTROLS
		//The images must have the same size
		if(!(img1.getWidth()==img2.getWidth() && img1.getWidth()==simg.getWidth()))
			throw new NotCompliantSizeException();
		if(!(img1.getHeight()==img2.getHeight() && img1.getHeight()==simg.getHeight()))
			throw new NotCompliantSizeException();
		//////////
		
		//create the output share images
		share_img1 = new BufferedImage(img1.getWidth()*2, img1.getHeight()*2, BufferedImage.TYPE_INT_ARGB);
		share_img2 = new BufferedImage(img2.getWidth()*2, img2.getHeight()*2, BufferedImage.TYPE_INT_ARGB);
	}
	
	
	/**
	 * Encode every pixel of innocent images to generate the share images.
	 * Property: after the original innocent looking images are encoded,
	 * they are still meaningful.  
	 */
	public void process()
	{	//first: from left to right
		//then: from up to down
		for(int i=0; i<simg.getHeight(); i++)
		{	for(int j=0; j<simg.getWidth(); j++)
			{	Color c = new Color(simg.getRGB(j,i));
				Color c1 = new Color(img1.getRGB(j,i));
				Color c2 = new Color(img2.getRGB(j,i));
				//encode this pixel for the shares
				encode(i, j, c, c1, c2);
			}
		}
	}
	
	/**
	 * Generate the share pixels and save them in the appropriated images
	 * @param i y-coordinate
	 * @param j x-coordinate
	 * @param c pixel color in the secret image
	 * @param c1 pixel color in the first image
	 * @param c2 pixel color in the second image
	 */
	public void encode(int i, int j, Color c, Color c1, Color c2)
	{	//preparing a vector with the permutation of {1...m}
		//it will contain a random permutation {i1, i2, ..., im}
		//we use it to generate an encoding matrix C that is a column permutation
		//of S 
		int permutation[] = getPermutation();
		
		//calculate index of the S basis matrix corresponding to c1,c2,c
		//every row i of S is an encoding of ci used to obtain c stacking
		//together the ci pixels
		int index = color2index(c, c1, c2);
		//obtain the "shares" pixels from a column permutation of this S matrix 
		int[][] shares = S[index].share(permutation);
		
		//now write encoding result to share images
		/* We use a 2x2 matrix to encode the m=4 expansion of every pixel.
		 * The "shares" pixel vector is transfered to the matrix by this order:
		 * [v(0) v(1) v(2) v(3)]  ==>	[v(0) v(1)]
		 *								[v(2) v(3)]
		 * e.g. [1 0 0 1] ==>	[1 0]
		 *						[0 1]
		 * In share images, (i,j) is a virtual pixel. Infact we consider his subpixels.
		 */ 
		for(int k=0; k<2; k++)
		{	for(int l=0; l<2; l++)
			{	//reconstructing RGB value of "shares" pixel colours
				//1->black  0->white
				int cs1 = (shares[0][k*2+l]==1)?Color.black.getRGB():Color.white.getRGB();
				int cs2 = (shares[1][k*2+l]==1)?Color.black.getRGB():Color.white.getRGB();
				
				//adapting loop variable to realize the transferring order
				int x=l;
				//set pixel color 
				share_img1.setRGB(2*j+x, 2*i+k, cs1);
				share_img2.setRGB(2*j+x, 2*i+k, cs2);
			}
		}
	}
	
	/** Evalueate a vector of length m with a random permutation
	 * of {1...m}. Let it be {i1, i2, ... , im}
	 * @param p the vector to evalueate
	 * @return the m-vector with the permutation
	 */
	public int[] getPermutation()
	{	//boolean vector for reminding just generated values
		boolean[] chk = new boolean[m];
		//integer vector to evalueate
		int[] p = new int[m];
		int j=0;
		while(j<p.length)
		{	int k = (int)(Math.random()*p.length);
			//if not just generated
			if(!chk[k])
			{	chk[k] = true; //mark it as "generated"
				p[j]=k; //add to p
				j++; //go on
			}
		}
		return p;
	}
	
	/**
	 * Let 3 colors generate the index for matrix by this law:
	 * let 1=black, 0=white
	 * index = c1*4+c2*2+c*1
	 * that is as a binary number.
	 * @return the index
	 */
	public int color2index(Color c, Color c1, Color c2)
	{	int a[] = new int[3];
		a[0]=(c1.getRGB()==Color.black.getRGB())?1:0;
		a[1]=(c2.getRGB()==Color.black.getRGB())?1:0;
		a[2]=(c.getRGB()==Color.black.getRGB())?1:0;
		int i = a[0]*4+a[1]*2+a[2]*1;
		return i;
	}
	
	/**
	 * Load basis matrices for the 2 out of 2 threshold EVCS
	 * The collection C(c,c1,c2) are obtained by permutating
	 * the columns of the respective matrix S(c,c1,c2).
	*/
	public void populateBasisMatrices()
	{	int[][] S000 ={ {1, 0, 0, 1},
						{1, 0, 1, 0}
					};
		S[0] = new BasisMatrix(S000);
		
		int[][] S001 ={ {1, 0, 0, 1},
						{0, 1, 1, 0}
					};
		S[1] = new BasisMatrix(S001);
		
		int[][] S010 ={ {1, 0, 0, 1},
						{1, 0, 1, 1}
					};
		S[2] = new BasisMatrix(S010);
		
		int[][] S011 ={ {1, 0, 0, 1},
						{0, 1, 1, 1}
					};
		S[3] = new BasisMatrix(S011);
		
		int[][] S100 ={ {1, 0, 1, 1},
						{1, 0, 1, 0}
					};
		S[4] = new BasisMatrix(S100);
		
		int[][] S101 ={ {1, 0, 1, 1},
						{0, 1, 1, 0}
					};
		S[5] = new BasisMatrix(S101);
		
		int[][] S110 ={ {1, 0, 1, 1},
						{1, 0, 1, 1}
					};
		S[6] = new BasisMatrix(S110);
		
		int[][] S111 ={ {1, 0, 1, 1},
						{0, 1, 1, 1}
					};
		S[7] = new BasisMatrix(S111);
	}
	
	/**
	 * Write the share images into the file f1, f2
	 * @param f1 file in wich save the 1st share image
	 * @param f2 file in wich save the 2nd share image
	 */
	public void save(File f1, File f2)
	{	try
		{	ImageIO.write(share_img1, "PNG", f1);
			ImageIO.write(share_img2, "PNG", f2);
		}
		catch(IOException e)
		{	System.out.println("Error, writing share images.");
			System.exit(1);
		}
	}
	
	/**
	 * @return an array with the share images
	*/
	public BufferedImage[] getShares()
	{	BufferedImage[] shareims = new BufferedImage[2];
		shareims[0] = share_img1;
		shareims[1] = share_img2;
		return shareims;
	}
}
